/* The "f9" object, by Miller Puckette */
/* copyright 1994 IRCAM */

/*
  This is a score follower that runs on two timescales.  Provisional
  pitches, on the right, can trigger outputs only if the follower
  is already up to the previous note.  "real", confirmed pitches, go in
  the left. They go to a Dannenburg-style score follower.
  
  Both provisional and confirmed pitches are floating-point.
  
*/

#define MAX_EXTERN
#include <m_extern.h>
#include <x_explode.h>
#define NNOTES 32

typedef struct _note
{
  t_evt *n_evt;
  struct _note *n_next;
  float n_value;
  long n_matchtime;
} t_note;

typedef struct _f9
{
  t_object x_ob;		/* object header */
  void *x_outlet5;
  void *x_outlet4;
  void *x_outlet3;
  void *x_outlet2;
  void *x_outlet1;

  char x_running;		/* true if we're turned on */
  char x_atend;			/* true if we've seen the last event */
  char x_spoofed;		/* true if we've output a speculative event */
  t_symbol *x_sym;		/* name of explode object */
  long x_serial;		/* last serial number in explode */
  t_note x_notes[NNOTES];
  t_note *x_tail;
  t_note *x_head;
  t_note *x_lastoutput;		/* last event that has been output */
  t_evt *x_firstev;
  float x_hit_score_in_tune;	/* parameters */
  float x_pitch_accuracy;
} t_f9;

void f9_stop(t_f9 *x);

t_symbol *f9_explodesym;

t_evt *f9_getfirstevent(t_f9 *x)
{
  t_explode *y;

  y = (t_explode *) (x->x_sym->s_thing);
  if ((!y) || (obj_getclassname(y) != f9_explodesym) )
  {
    error( "f9 : %s is not an explode", x->x_sym->s_name);
    return (0);
  }
  if (y->x_next)
    warn( "f9 : %s is multiply defined", x->x_sym->s_name);
  if (!y->x_serial)
  {
    error("f9 : %s is recording", x->x_sym->s_name);
    return (0);
  }
  if (x->x_serial && x->x_serial != y->x_serial)
  {
    error("f9 : %s was changed while following was on", x->x_sym->s_name);
    return (0);
  }
  x->x_serial = y->x_serial;
  return (y->x_evt);
}

void f9_refill(t_f9 *x)
{
  t_evt *e;
  t_note *note = x->x_head->n_next;
  float value = x->x_head->n_value;
  int matchtime = gettime() - 10000;
  if (x->x_atend) return;

  e = x->x_head->n_evt->e_next;
  while (e && note != x->x_tail)
  {
    if (e->e_chan == 1)
    {
      value -= e->e_vel;
      note->n_evt = e;
      note->n_value = value;
      x->x_head = note;
      note = note->n_next;
    }
    e = e->e_next;
  }
  if (!e) x->x_atend = 1;
}

void f9_hit_score_in_tune(t_f9 *x, double f) {
  x->x_hit_score_in_tune = f; }

void f9_pitch_accuracy(t_f9 *x, double f) {
  x->x_pitch_accuracy = f; }

void f9_print(t_f9 *x)
{
  t_note *n;
  post("f9: running %d, atend %d, spoofed %d", x->x_running, x->x_atend,
    x->x_spoofed);
  post("notes %lx, lastoutput %lx", x->x_notes, x->x_lastoutput);
  if (x->x_running ) for (n = x->x_tail; 1; n = n->n_next)
  {
    if (n->n_evt)
      post("%d %9.2f",
	n->n_evt->e_pit, n->n_value);
    else post("(0)");
    if (n == x->x_head) break;
  }
  post("hit-score-in-tune %f", x->x_hit_score_in_tune);
  post("pitch-accuracy %f", x->x_pitch_accuracy);
}

void f9_sendevt(t_f9 *x, t_evt *e, int noteon)
{
  outlet_int(x->x_outlet5, e->e_chan);
  outlet_int(x->x_outlet4, e->e_dur);
  outlet_int(x->x_outlet3, noteon ? e->e_vel : 0L);
  outlet_int(x->x_outlet2, e->e_pit);
}

void f9_stop(t_f9 *x)
{
  x->x_running = 0;
  x->x_serial = 0;
}

  /* advance to note pointed to by "newnote" */ 
void f9_advanceto(t_f9 *x, t_note *newnote)
{
  t_evt *e;
  t_note *wanttail;
    /* output notes on channels other than 1 */
  for (e = (x->x_lastoutput ? x->x_lastoutput->n_evt : x->x_firstev);
      e != newnote->n_evt; e = e->e_next)
  {
    if (e->e_chan != 1) f9_sendevt(x, e, 1);
  }
    /* send the matched note (helps monitor progress) */
  f9_sendevt(x, e, 1);
    /* update value of "lastoutput" */
  x->x_lastoutput = newnote;

  if (x->x_atend)
  {
    if (newnote == x->x_head)
    {
      post("end 1");
      f9_stop(x);
    }
  }
  else
  {
      /* scoop up more notes for score follower */
    wanttail = newnote + (NNOTES/2);
    if (wanttail > x->x_tail + NNOTES) wanttail -= NNOTES;
    if (wanttail < x->x_tail + NNOTES/2 && wanttail > x->x_tail)
    {
      x->x_tail = (wanttail > x->x_notes + NNOTES ?
	wanttail - NNOTES : wanttail);
      f9_refill(x);
    }
  }
}

void f9_note(t_f9 *x, double f, double extra_note_penalty_in_tune,
 double extra_note_penalty_out_of_tune)
{
  float pitch = f, lastvaluebefore = 0, lastvalueafter = 0,
    bestvalue = -1000;
  t_note *current, *end, *bestnote;
  int hitlast = 0, goforward = 0;
  float ACCURACY = (x->x_pitch_accuracy > .001 ? x->x_pitch_accuracy : .001);
  float recip = 1./ACCURACY;
  float MATCHVALUE = x->x_hit_score_in_tune;
  float EXTRAVALUE = -extra_note_penalty_in_tune;
  float EXTRADETUNE =
    ((float)(extra_note_penalty_in_tune -
      extra_note_penalty_out_of_tune)) * recip;

  if (!x->x_running) return;
  if (!f9_getfirstevent(x))
  {
    f9_stop(x);
    return;
  }
  
  bestnote = NIL;
  for (current = x->x_tail, end = x->x_head; 1;
    current = current->n_next)
  {
    float deviation = pitch - current->n_evt->e_pit, v1, v2, v3;
    if (deviation < 0) deviation = -deviation;
    if (deviation > ACCURACY) deviation = ACCURACY;
    v1 = current->n_value + EXTRAVALUE + EXTRADETUNE * deviation;
    v2 =  lastvaluebefore + MATCHVALUE -
      (current->n_evt->e_vel + MATCHVALUE) * deviation * recip;
    v3 = lastvalueafter - current->n_evt->e_vel;
    
    lastvaluebefore = current->n_value;
    lastvalueafter = v1;
    if (v2 > lastvalueafter) lastvalueafter = v2;
    if (v3 > lastvalueafter) lastvalueafter = v3;
    current->n_value = lastvalueafter;
    if (lastvalueafter > bestvalue)
    {
      bestvalue = lastvalueafter;
      bestnote = current;
      if (hitlast || (!x->x_lastoutput && bestvalue > 0.)) goforward = 1;
    }
    if (current == x->x_lastoutput) hitlast = 1;
    if (current == end) break;
  }

  if (goforward)
  {
    f9_advanceto(x, bestnote);
    x->x_spoofed = 0;
  }
  else if (x->x_spoofed && bestnote == x->x_lastoutput) x->x_spoofed = 0;
}

void f9_float(t_f9 *x, double f)
{
  t_note *note;
  double deviation;
  if (!x->x_running || x->x_spoofed || !f9_getfirstevent(x) ||
    x->x_lastoutput == x->x_head) return;
  note = (x->x_lastoutput ? x->x_lastoutput->n_next : x->x_tail);
  deviation = note->n_evt->e_pit - f;
  if (deviation < 0.) deviation = -deviation;
  if (deviation <= .5 * x->x_pitch_accuracy)
  {
    f9_advanceto(x, note);
    x->x_spoofed = 1;
  }
}

void f9_startup(t_f9 *x, t_evt *e)
{
  x->x_atend = 0;
  x->x_firstev = e;
  while (e && e->e_chan != 1) e = e->e_next;
  if (!e)
  {
    f9_stop(x);
    return;
  }
  x->x_notes[0].n_evt = e;
  x->x_notes[0].n_value = - e->e_vel;
  x->x_notes[0].n_matchtime = gettime() - 10000;
  x->x_head = x->x_tail = x->x_notes;
  x->x_lastoutput = 0;
  x->x_spoofed = 0;
  f9_refill(x);
  x->x_running = 1;
}

/* ---------- method follow : start following ---------- */
void f9_follow(t_f9 *x)
{
  t_evt *e;
  f9_stop(x);
  e = f9_getfirstevent(x);
  if (e) f9_startup(x, e);
}

/* ---------- method followat : search for a note and start there ---------- */
void f9_followat(t_f9 *x, long pitch, long velocity, long chan)
{
  t_evt *e;
  f9_stop(x);
  e = f9_getfirstevent(x);
  if (e)
  {
    for (; e; e = e->e_next)
      if (e->e_pit == pitch && e->e_vel == velocity && e->e_chan == chan)
	break;
    if (!e) post("f9: note not found");
    if (e) e = e->e_next;
  }
  if (e) f9_startup(x, e);
}

/* ---------- method params : set the params for score following ---------- */
void f9_params(t_f9 *x, long nfwd, long ftime, long oct, long tolerance)
{
}

/* ---------- method free : free the object ---------- */
void f9_free(t_f9 *x)
{
  /* should free the hang and skip lists */
}

/* ---------- pointer to the class ---------- */
t_externclass *f9_class;

/* ---------- method new : instance creation ---------- */
void *f9_new(t_symbol *s)
{
  t_f9 *x;
  t_explode *y;
  t_note *n;
  int i;

  x = (t_f9 *) obj_new(&f9_class, 0);

  /* create outlet, declaring it to send "int" */
  x->x_outlet5 = outlet_new(x, gensym("int"));
  x->x_outlet4 = outlet_new(x, gensym("int"));
  x->x_outlet3 = outlet_new(x, gensym("int"));
  x->x_outlet2 = outlet_new(x, gensym("int"));
  x->x_outlet1 = outlet_new(x, gensym("int"));

  x->x_sym = s;   /* initialise symbol to connect to explode object */

  for (i = 0, n = x->x_notes; i < NNOTES; i++, n++)
    n->n_next = (i == NNOTES-1 ? x->x_notes : n+1);

  f9_hit_score_in_tune(x, 5.);
  f9_pitch_accuracy(x, 1.);

  f9_stop(x);
  return x;
}

/* ---------- method setup : class creation ---------- */
f9_setup()
{
  c_extern( &f9_class, f9_new, f9_free,
    gensym("f9"), sizeof(t_f9), 0, A_SYM, 0);

  c_addmess(f9_note, gensym("note"), A_FLOAT, A_FLOAT, A_FLOAT);

  c_addfloat(f9_float);

  c_addmess(f9_follow, gensym("follow"), 0);
  c_addmess(f9_followat, gensym("followat"),
		A_LONG, A_LONG, A_LONG, 0);

  c_addmess(f9_stop, gensym("stop"), 0);

  c_addmess(f9_print, gensym("print"), 0);

  c_addmess(f9_hit_score_in_tune,
    gensym( "hit-score-in-tune"), A_FLOAT, 0);
  c_addmess(f9_pitch_accuracy,
    gensym( "pitch-accuracy"), A_FLOAT, 0);

  f9_explodesym = gensym("explode");
}

